
package sam.infer ; 

import java.util.* ; 
import sam.lang.* ; 
import sam.util.* ; 


/**
  *  The DC implementation of ActiveRule
  *  The distributive choice algorithm provides 
  *  incremental saturation for choices on the branch
  *  without retry.  The algorithm is FAIR: 
  *  it eventually attempts all choices possible. 
  *    @version 16 August 2014 (modified 10/8/2023)
  *    @author &copy; John Fisher
   */
public class ActiveDCRule extends ActiveRule { 
  
   // state and range parameters
   protected int[] top , bottom ;  // current range limits
   protected int trv[],    // top of range vals
                 mrv[],    // middle range vals
                 brv[]  ;  // bottom of vals
   protected int state ;   // 0 .. A
   protected boolean start ;
   //protected int mark ;   
   protected boolean new_ranges_needed ; // set new ranges after states 0 & 2 
   protected boolean[] extended ;
   protected Stack<Extent> extentStack ; // for rollBack

   /**
     * Construct an ActiveDCRule 
      */
   public ActiveDCRule(Rule r,Tree dftree,ActiveTheory active,int indx) { 
      // r.A is size of the antecedent
      super(r,dftree,active,indx) ; 
      // current range parameters for hasMore()
      top =    new int[A] ;
      bottom = new int[A] ;
      // current state range values
      trv =    new int[A] ;  // top range vals
      mrv =    new int[A] ;  // middle range vals
      brv =    new int[A] ;  // bottom range vals
      extended = new boolean[A] ; 
      state = 0 ; 
      start = true ; 
      mark = -1 ; 
      extentStack = new Stack<Extent>() ; 
   }  

   /**
     *  Apply this ActiveDCRule to focus in the Tree 
      */
   public int apply() { // use facts to focus  
         //this.start = true ;  p1p2 TOO DEEP ...
         //System.out.println("rule#"+this.index+"@"+tree.count+":"+"U="+this.U+" mark="+this.mark+" start="+this.start) ;
      if (this.U == 0 && this.mark > -1) return 0 ;  
      if (start) {  // start state  
         // new_ranges_needed = false ;          
         // test for a start
         for (int i= 0 ; i < A ; i++) {
            trv[i] = activeTheory.first[FACTOR[i].index] ; 
            if (trv[i] == tree.depth) return 0 ;  // quick return 
         }
         // get here, have started
         start = false ; // stays false until reset
         for (int a = 0 ; a < A ; a++) extended[a] = false ;
         state = 0 ; 
         for (int i= 0 ; i < A ; i++) {
            brv[i] = activeTheory.last[FACTOR[i].index] ;
            top[i] = trv[i] ;
            bottom[i] = brv[i] ;
            choice[i] = trv[i] ;
            for (int u = 0 ; u < U ; u++) table[u][i] = null ; 
         }
         factor = 0 ; 
         mark = -1 ; 
         new_ranges_needed = false ; 
      }
      if (hasNext()) { // applicable 
         this.mark = tree.focus ;  // mark point of application 
         return expand() ;         // will update tree.splitsCount
         // expand checks bounds and complexity
      }
      else { // failure to have more choices - state change
         return 0 ;        // not applicable now
      }
   }


   /**

     * Set range extensions.
     * @return true if range extends, false otherwise
      */
   protected boolean setNewRanges() { 
      for (int a = 0 ; a < A ; a++) { 
            mrv[a] = brv[a] ; 
            brv[a] = activeTheory.last[FACTOR[a].index] ; 
            extended[a] = (brv[a] > mrv[a]) ;  // record whether extended
            for (int u = 0 ; u < U ; u++) table[u][a] = null ; // clear table
      }
      // any range extension?  
      for (int a = 0 ; a < A ; a++) 
         if(extended[a]) {  // new range for factor a 
            // save old ranges NOW
            extentStack.push(new Extent(this.mark,mrv)) ;
            return true ;  
         }
      return false ;    }
      /*
      */

   /**
     * Set current range choices for state (or bail)
     * Note that also when a factual factor extends, that makes the 
     * rule possibily applicable to new DC choices.  
     * @return true if state has choices, else false
      */
   protected boolean setChoices(int state) {  
      if (state > A) { 
         new_ranges_needed = true ; 
         return false ; 
      }
      int d = A-state ;
      // quick check
      if (!extended[d]) return false ; // state has no choices
      top[d] = activeTheory.jumpto[mrv[d]] ; // drop down  ???
      //if (top[d] == tree.depth) return false ;  //??? NO, since extended
      bottom[d] = brv[d] ;
      for (int i = 0 ; i < d ; i++) { 
         top[i] = trv[i] ;
         bottom[i] = mrv[i] ; 
      }
      for (int i = d+1 ; i < A ; i++) { 
         top[i] = trv[i] ;
         bottom[i] = brv[i] ; 
      }
      this.state = state ; 
      return true ;
   }

   /**
     *   Control states 1..A, ranges and call hasMore().
      */
   public boolean hasNext() { 
      // on the fly ...
      while (true) {
         if (new_ranges_needed) { 
            if (!setNewRanges()) return false ;  // no extended ranges now
              // maybe later, N.B. new_ranges_needed == true (still)
              // otherwise find usable state/choices
            new_ranges_needed = false ;
            state = 1 ;  
            // advance to a useful state
            while (!setChoices(state)) state++ ;
            for (int i = 0 ; i < A ; i++) choice[i] = top[i] ; // NOT here
            new_ranges_needed = false ;
            factor = 0 ;    
         }
         if (hasMore()) { 
            mark = tree.focus ; 
            return true ; 
         }
         else { // !hasMore() 
            state++ ; 
            // advance to a useful state
            while (state <= A && !setChoices(state)) state++ ;
            if (state == A+1) new_ranges_needed = true ; 
            else { 
               // go to tops of ranges
               for (int i = 0 ; i < A ; i++) choice[i] = top[i] ; 
               // clear table 
               for (int i= 0 ; i < A ; i++) 
                  for (int u = 0 ; u < U ; u++) table[u][i] = null ; 
               factor = 0 ;
            }
         }
      } // end while
   }

///// spy ///////////////////////////////////////////////////////
   // spy range
   private String indexrange(int indx) {
      StringBuffer b = new StringBuffer() ; 
      b.append("<") ; 
      int nxt = activeTheory.first[indx] ;
      while (nxt < tree.depth) {
         b.append(nxt) ; 
         nxt=activeTheory.jumpto[nxt] ;
         if (nxt < tree.depth)  b.append(",") ;
      }
      b.append(">") ;
      return b.toString() ; 
   }

   private void spyExtended() {
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for (int a = 0 ; a < A ; a++) {
         sb.append(extended[a]) ; 
         if (a < A-1) sb.append(",") ;
      }
      sb.append("]") ;
      System.out.println("extended="+sb.toString()) ;
   }

   // Display range of choices
   private void spyRanges(int indx) {
      if (this.index == indx) {
         System.out.println("focus="+tree.focus) ; 
         for(int i = 0 ; i < A ; i++) {
            System.out.println("top["+i+"]="+top[i]) ; 
            System.out.println("bottom["+i+"]="+bottom[i]) ; 
         }
      }
   }

   // Display choice
   protected void spyChoice(int indx) { 
      if (this.index == indx) {
         StringBuffer b = new StringBuffer() ; 
         b.append("@"+tree.focus+" rule#") ; 
         b.append(indx+":[") ; 
         for(int i = 0 ; i < U ; i++) { 
            b.append(String.valueOf(choice[i])) ;
            if (i < U-1) b.append(",") ;
         }
         b.append("]") ; 
         System.out.println( b.toString() ) ;
      }
   }
////////////////////////////////////////////////////////////

   /**
     *  Linear search current ranges for new matches.  
     *  @return true on new match, false otherwise
     *  Note top & bottom range guards: need to stay in bounds
      */
   protected boolean hasMore() { 
      if (factor == A) {  
         //initial = false ;
         factor = A-1 ;    // last actual factor
         choice[factor] = activeTheory.jumpto[choice[factor]] ; 
         // nullify factor bindings 
         for (int i = 0 ; i < U ; i++) 
            table[i][factor] = null ;
      }
      while (factor != -1) {   // until fails to have next
         if ( factor == A ) {  // all factors satisfied  
            if (!subsumed()) { 
               return true ;  // new state 
            }
            else { // subsumed so backtrack 
               factor = A-1 ;    // last actual factor
               choice[factor] = activeTheory.jumpto[choice[factor]] ;  
                                              // may be out of bounds
               // nullify factor bindings 
               for (int i = 0 ; i < U ; i++) 
                  table[i][factor] = null ;
            }
         }
         else { 
            Functor f = FACTOR[factor] ;
            int p = choice[factor] ;   
            // travel down tree branch to attempt match
            boolean matched = false ; 
            // avoid calling match for different functors
            while (p <= bottom[factor]) {   // stay in bounds  
                  //Fact fxd = tree.branch[p] ;
                  //boolean same = f.index == fxd.index ; 
                  //if (!same) { 
                  //   System.out.println("count="+tree.count+": rule#"+this.index+" free="+f+" index="+f.index +" range="+indexrange(8)+" for p="+p + " fxd=" + fxd.toString(this.activeTheory.theory)+" index="+fxd.index) ; 
                  //}
                  //if (same && match_order_args(factor,f,fxd)) {  
                  if ( match_order_args(factor,f,tree.branch[p])) {  
                                                // match arguments? 
                     matched = true ;
                     choice[factor] = p ; // where matched in the tree  
                     break ;
                  }
                  else {  
                     // undo any partial bindings
                     // nullify factor column 
                     for(int v = 0 ; v < U ; v++) 
                        table[v][factor] = null ;     
                     ///// down tree
                     p = activeTheory.jumpto[p] ;   
                  }
            }  // end while
            // test result of attempt(s) to match
            if (matched) factor++ ; // next factor
            else { // no match for factor 
               factor-- ; 
               if (factor > -1) { // requires A > 1, 
                  choice[factor+1] = top[factor+1] ; 
                  // clear bindings
                  for(int u = 0 ; u < U ; u++)  
                     table[u][factor+1] = null ;  
                  // adjust factor 
                  choice[factor] = activeTheory.jumpto[choice[factor]] ;  
                  // nullify bindings for factor
                  for(int u = 0 ; u < U ; u++)  
                     table[u][factor] = null ;  
               }
               // otherwise factor == -1 and the loop stops
            }  
         }
      } 
      return false ; // nothing worked this attempt
   }

///// rollBack(-) redux all over again

   // ALWAYS restart DC at branch split
   public void rollBackVER1(int p) { // v1, rename reverts
      this.start = true ;
      this.mark = -1 ;  
      //for (int a = 0 ; a < A ; a++) extended[a] = false ;
   }  // p1p2, 42ms

   // rollback to beginning only if some data > p
   public void rollBackVER2(int p) { //v2, rename reverts OK
      boolean warn = false ; 
      // warn if inference below branch point 
      // choices might not have been
      if (this.mark > p) warn = true ; 
      // warn if any range extends below branch point,
      // say failed to fire below p
      for (int i = 0 ; i < A ; i++) 
         if (brv[i] > p) { warn = true ; break ; }      
      if (warn) {  // restart from root on branch
         this.start = true ;
         this.mark = -1 ; 
         for (int a = 0 ; a < A ; a++) extended[a] = false ;
      } // otherwise continue in current state
   } // p1p2, 38ms

   /* 
    * Use the extentStack to find calculated ranges.
    * @version 10/8/2023 -- Quandry modification:
    * Set start=true for each new branch.
   */
   public void rollBack(int p) { //v3, rename reverts  
      //  Check whether curent ranges are OK
      ///// just like v2 ///////////////////////
      boolean warn = false ; 
      if (this.mark > p) warn = true ; 
      for (int i = 0 ; i < A ; i++) 
         if (brv[i] > p) { warn = true ; break ;}
      ///////////////////////////////////////////
      if (warn) {  // look for better ranges
         Extent xtnt = null ; 
         while (!extentStack.isEmpty()) { 
            xtnt = extentStack.pop() ; 
            if (!xtnt.below(p)) break ; // usable
            else xtnt = null ;
         }  // using null value as fail test
         if (xtnt == null) {   
            this.start = true ;
            this.mark = -1 ; 
         }
         else {  // align ranges and mark with xtnt
            for (int a = 0 ; a < A ; a++) this.brv[a] = xtnt.brv[a] ; 
            this.mark = xtnt.mark ;
            for (int a = 0 ; a < A ; a++) extended[a] = false ;
            this.new_ranges_needed = true ;
         }
      } 
      // 10/8/2023 for switch to new newbranch, see explanation at      
      // https://skolemmachines.org/related/order_quandry/order_quandry.html
      this.start = true ; 
   }   

   // inner class, record extent of ranges and marks for rollBack
   class Extent {
      int mark ;   // last inference
      int[] brv ;  // bottom of ranges

      Extent(int m, int[] brv) { 
         this.mark = m ;
         this.brv = new int[A] ;
         for (int i=0 ; i < A ; i++) this.brv[i]=brv[i] ; 
      }
 
      boolean below(int p) { 
         if (this.mark > p) return true ;
         for (int i=0 ; i < A ; i++)
           if (this.brv[i] > p) return true ;
         return false ;
      }
   }  

   private static String show(int[] c) { 
      StringBuffer b = new StringBuffer() ; 
      b.append("[") ;
      for (int i = 0 ; i < c.length ; i++) {
         b.append(c[i]) ; if (i < c.length -1) b.append(",") ; 
      }
      b.append("]") ; 
      return b.toString() ; 
   }
}

//////////////////////////////////////////////////////////////////
/////   in apply ...
/////  
/////      this.mark = tree.focus ; // before  expand()
/////      return expand() ;
/////
/////   marks the point of application: 
/////
/////         |
/////         .  tree.focus (now) == this.mark
/////         |
/////         |   ...
/////         |
/////         p  tree.focus (after expand)
/////
/////   If p is a branch point, it is required 
/////   that  mark < p in order for this rule's 
/////   expansion to lie entirely above that branch point.
/////   Each brv[i] < p means that no range extends 
/////   below p, thus that the ranges are good for the next
/////   branch also.  Notice that there might be ranges extending
/////   below mark that have been considered but not expanded, in 
/////   which case restart is needed.
/////   Also it could be that each brv[i] is above p
/////   but mark is at or below p, in which case restart is needed.
/////   Need brv[i] < p (not <=) is demonstrated by dpe.eq.co
/////////////////////////////////////////////////////////////////////

