package sam.infer ; 

import sam.lang.* ; 
import sam.util.* ; 
import java.util.* ; 

/**
  *  An ActiveTheory maintains lists of ActiveRules
  *  associated with a colog Theory, using DC rules.
  *  @version 26 June 2012, using Tree.table
  *  @author &copy; John Fisher
  */ 
public class ActiveTheory {

   /**
     *  The sam.lang.Theory associated with this ActiveTheory.
     */
   public Theory theory ; 
   
   /**
     *  The "definite" rules of the theory.
     *  The rules that are NOT queued.
     */
   public ActiveDCRule[] dRule ; 

   /**
     *  The "queued" rules of the theory.
     */
   public ActiveDCRule[] qRule ; 

   /**
     *  The Tree associated with this ActiveTheory.
     */
   public Tree tree ; 

   /*** RoundRobin of existential rule #'s  */
   protected RoundRobin rrQ ; 

   /*** Q'ed or not */
   private boolean[] qued ; 

   /*** jumpto next predicate occurrence on tree branch */
   protected int[] jumpto ;    // jumpto[i] = j means next occurrence of predicate at branch[i] is at branch[j]
   /*** first occurrence of predicate on branch */
   protected int[] first ;    // first[i]=j means fact having name[i] has first occurrence at branch[j]
   /*** last occurrence of predicate on branch */
   protected int[] last ;    // last[i]=j means fact having name[i] has last occurrence at branch[j]

   public Rule removed ; 

   /**
     *  The ActiveTheory determine by a colog Theory
     *  and associated with a Tree.
     *  Manual queueing.
     */
   public ActiveTheory(Theory thy, Tree t) { 
      this.theory = thy ; 
      //////////
      CologAnalyzer.analyze(thy) ; 
      //////////
      // MANUAL queueing option $do => queue(i,j,k,...). as first 0th rule 
      Functor dofunc = new Functor("$do",0,Functor.FCNL) ; 
      Rule rule0 = thy.elementAt(0) ; 
      // test for do=>queue(-)  manual queueing experiment
      if (rule0.antecedent.size()==1 &&
            rule0.antecedent.elementAt(0).equals(dofunc) &&
            rule0.consequent.size()==1 &&
            rule0.consequent.elementAt(0).size()==1 &&
            rule0.consequent.elementAt(0).elementAt(0).name.equals("queue")) {  
         // manual queueing 
         Term[] args = rule0.consequent.elementAt(0).elementAt(0).args ;
         Vector<Integer> queued = new Vector<Integer>() ;
         for (int i=0 ; i < args.length ; i++) 
            queued.add( Integer.parseInt((String)((Functor)args[i]).name)) ;
         this.removed = thy.remove(0) ; // lose "do=> ..."
         this.theory = thy ; 
         this.qued = new boolean[thy.size()] ;
         this.tree = t ; 
         int nq ;
         if (queued.contains(0)) nq = 0 ;
         else nq = queued.size() ;  // how big for qRule
         dRule = new ActiveDCRule[thy.size() - nq] ; 
         qRule = new ActiveDCRule[nq] ;
         int d = 0, q = 0 ;
         for (int i=0 ; i < thy.size() ; i++) {
            Rule r = (Rule)thy.elementAt(i) ; 
            if (queued.contains(i+1)) {
               if (r.factual || r.subfactual) qRule[q] = new ActiveTabDCRule(r,t,this,i+1) ;  
               else qRule[q] = new ActiveDCRule(r,t,this,i+1) ;  
               qued[i] = true ; 
               q++ ; 
            }   
            else { 
               if (r.factual || r.subfactual) dRule[d] = new ActiveTabDCRule(r,t,this,i+1) ;  
               else dRule[d] = new ActiveDCRule(r,t,this,i+1) ;  
               d++ ; 
            }
         }
         rrQ = new RoundRobin(nq) ; // just big enough & next is 0
      }
      else {  // automatic queueing
         this.theory = thy ; 
         this.qued = new boolean[thy.size()] ;
         this.tree = t ; 
         // compute number of expansive rules to queue
         Iterator it = thy.iterator() ; 
         int nq = 0 ;  // how big for qRule
         while (it.hasNext()) { 
            Rule r = (Rule)it.next() ;
            if (r.existential > 0 || r.complex ) 
               nq++ ; 
         }
         dRule = new ActiveDCRule[thy.size() - nq] ; 
         qRule = new ActiveDCRule[nq] ;
         int d = 0, q = 0 ;
         for (int i = 0 ; i < thy.size() ; i++) {
            Rule r = (Rule)thy.elementAt(i) ; 
            if (r.existential > 0 || r.complex) {   // expansive?
               if (r.factual || r.subfactual) qRule[q] = new ActiveTabDCRule(r,t,this,i+1) ;  
               else qRule[q] = new ActiveDCRule(r,t,this,i+1) ;  
               qued[i] = true ; 
               q++ ; 
            }   
            else { 
               if (r.factual || r.subfactual) dRule[d] = new ActiveTabDCRule(r,t,this,i+1) ;  
               else dRule[d] = new ActiveDCRule(r,t,this,i+1) ;  
               d++ ; 
            }
         }
         rrQ = new RoundRobin(nq) ; // just big enough
      } // end automatic queueing
      // initialize indexing for tree
      this.jumpto = new int[this.tree.depth] ; 
      // no next occurrences yet ( jumpto[-] > focus ) 
      for (int i = 0 ; i < tree.depth ; i++) 
         this.jumpto[i] = tree.depth ; 
      this.first = new int[theory.name.length] ;
      // no first yet ( first[-] > focus )
      for (int i = 0 ; i < first.length ; i++) 
         this.first[i] = tree.depth ; 
      // except true
      this.first[Theory.TRUE_INDEX] = 0 ; 
      this.last = new int[theory.name.length] ;  
      // no last yet 
      for (int i = 0 ; i < last.length ; i++) 
         this.last[i] = tree.depth ; 
      // except true
      this.last[Theory.TRUE_INDEX] = 0 ; 
   }

   /**
     * Tell others whether rule is queued (1-based) 
      */
   public boolean qued(int q) { return this.qued[q-1] ; }

   /**
     *  Apply this ActiveTheory to focus.
     */
   public int apply() { 
      if (tree.branch[tree.focus].index==Theory.GOAL_INDEX ||
            tree.branch[tree.focus].index==Theory.FALSE_INDEX ) return tree.focus ;
      int probe ;
      boolean flag = false ;
      //  The while loop makes sure that dRules are exhausted.
      //System.out.println("") ;
      while(true) {  
         for (int r = 0 ; r < dRule.length ; r++) {
            //System.out.print(r+" ") ; // which rule?
            probe = dRule[r].apply() ; 
            /////// probe ///////////////////////////
            // -2  rule applies but blocked by width
            // -1  rule applies but blocked by depth 
            //  0  this rule does not apply*
            //  n  applies and returns new focus 
            ////////////////////////////////////////
            if (probe != 0) return probe ; 
         }
         // Get here means no definite rule worked, so try next qRules 
         for (int c = 0 ; c <= rrQ.last ; c++) {  // visit all if necessary 
            int k = rrQ.next() ; // circular 
            probe = qRule[k].apply() ;  
            if (probe != 0) return probe ;  // see note above
         }
         // ELIMINATE THIS EXTRA EFFORT IF POSSIBLE
         // Get here, no rules worked, potential counter model.
         // Can make sure dRule(s) are exhausted.
         if (flag) return 0 ; 
         else flag = true ; // try again, may have earlier matches
         //return 0 ; 
      } // end while 
   }

   /**
     *  Reset the last-used marks. 
     *  Called via Tree.dfSearch(), on switch to new branch
     */
   protected void rollBack(int p) { 
      for (int r = 0 ; r < dRule.length ; r++) 
         dRule[r].rollBack(p) ; 
      for (int r = 0 ; r < qRule.length ; r++)  
         qRule[r].rollBack(p); 
   }


}






